##########################################################################
# Hopla - Copyright (C) AGrigis, 2015 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module contains a 'hopla' function used to create a list of commands
that will be executed in parallel using the local machine or cluster with
TORQUE resource manager installed. For the moment, the input script must be
a Python script.
"""  # pragma: no cover

# System import
import os  # pragma: no cover
from collections import OrderedDict  # pragma: no cover

# Hopla import
from .scheduler import scheduler  # pragma: no cover


def hopla(python_script, hopla_outputdir=None, hopla_cpus=1,
          hopla_use_subprocess=False, hopla_delay_upto=0, hopla_optional=None,
          hopla_append_kwargs=None, hopla_logfile=None, hopla_verbose=1,
          hopla_cluster=False, hopla_cluster_logdir=None,
          hopla_cluster_queue=None, hopla_cluster_memory=1,
          hopla_cluster_walltime=72, hopla_cluster_nb_threads=1,
          hopla_python_cmd="python", hopla_iterative_kwargs=None,
          **kwargs):
    """ Execute a python script/file in parallel.

    Rules:

        * This procedure enables local or cluster runs.
        * This procedure returns a human readable log.
        * In the command line, prefix kwargs with a '-' if name not in
          'hopla_optional' list, otherwise with a '--'.
        * In order not to interfer with command line kwargs use 'hopla' prefix
          in function parameters.
        * If the script contains a '__hopla__' list of parameter names to keep
          trace on, all the specified parameters values are stored in the
          execution status.
        * On a cluster if the 'hopla_python_cmd' parameter is set to
          None, execute directly the script as a command. Usefull for
          singularity call.

    Parameters
    ----------
    python_script: str (mandatory)
        a python script or file to be executed.
    hopla_outputdir: str (optional, default None)
        a folder where synthetic results are written.
    hopla_cpus: int (optional, default 1)
        the number of cpus to be used.
    hopla_use_subprocess: bool, default False
        use a subprocess (for instance in case of memory leak). In this
        particular case the __hopla__ setting is deactivated.
    hopla_delay_upto: int (optional, default 0)
        the processes' execution will be delayed randomly by [0, <delay_upto>[
        seconds.
    hopla_optional: list of str (optional, default None)
        the optional kwargs names that will be prefixed with '--' in the
        command line.
    hopla_append_kwargs: list of str (optional, default None)
        this is useful to allow the kwargs names to be specified multiple
        times.
    hopla_logfile: str (optional, default None)
        location where the log messages are redirected: INFO and DEBUG.
    hopla_verbose: int (optional, default 1)
        0 - display no log in console,
        1 - display information log in console,
        2 - display debug log in console.
    hopla_cluster: bool (optional, default False)
        if True use a worker that submits the jobs to a cluster.
    hopla_cluster_logdir: str (optional, default None)
        an existing path where the cluster error and output files will be
        stored. This folder must be empty.
    hopla_cluster_queue: str (optional, default None)
        the name of the queue where the jobs will be submited.
    hopla_cluster_memory: float (optional, default 1)
        the memory allocated to each job submitted on a cluster (in GB).
    hopla_cluster_walltime: int (optional, default 72)
        the walltime used for each job submitted on the cluster (in hours).
    hopla_cluster_nb_threads: int (optional, default 1)
        the number of cores allocated for each node.
    hopla_python_cmd: str (optional, default 'python')
        the path to the python binary. If None consider the command directly.
    hopla_iterative_kwargs: list of str (optional, default None)
        the iterative script parameters.
    kwargs: dict (optional)
        the script parameters: iterative kwargs must contain a list of elements
        and must all have the same length, non-iterative kwargs will be
        replicated.

    Returns
    -------
    execution_status: dict
        a dictionary that contains all the executed command return codes.
    exitcodes: dict
        a dictionary with a summary of the executed jobs exit codes.
    """
    # Function parameters
    hopla_optional = hopla_optional or []

    # Create the commands to be executed by the scheduler
    iterative_kwargs = hopla_iterative_kwargs or []
    append_kwargs = hopla_append_kwargs or []
    commands = []
    values_count = []
    # > sort kwargs
    kwargs = OrderedDict(sorted(list(kwargs.items())))
    # > deal kwargs with iterative kwargs
    for name, values in kwargs.items():
        if name in iterative_kwargs:
            if not isinstance(values, list):
                raise ValueError(
                    "All the iterative kwargs must be of list type. Parameter "
                    "'{0}' with value '{1}' does not fulfill this "
                    "rule.".format(name, values))
            values_count.append(len(values))
            for index, val in enumerate(values):
                if len(commands) <= index:
                    commands.append([])
                if val is None:
                    continue
                # > in the command line, prefix kwargs with '-',
                # optional kwargs with '--'
                if name in hopla_optional:
                    option = ["--" + name]
                else:
                    option = ["-" + name]
                # > set the option value
                if name in append_kwargs:
                    _option_key = option[0]
                    option = []
                    for elem in val:
                        option.append(_option_key)
                        if isinstance(elem, list):
                            option.extend(elem)
                        else:
                            option.append(elem)
                elif isinstance(val, list):
                    option.extend([str(item) for item in val])
                elif not isinstance(val, bool):
                    option.append(str(val))
                # > set the option in the command line: for boolean set the
                # option only if the value is true
                if isinstance(val, bool):
                    if val:
                        commands[index].extend(option)
                else:
                    commands[index].extend(option)
    # > check iterative kwargs values have the same length
    if (values_count != [] and
            values_count.count(values_count[0]) != len(values_count)):
        raise ValueError("All the iterative kwargs must have the "
                         "same number of values in order to iterate.")
    # > add script name to command line and deal with non iterative kwargs
    for cmd in commands:
        cmd.insert(0, python_script)
        for name, value in kwargs.items():
            if name not in iterative_kwargs:
                # > skip None value
                if value is None:
                    continue
                # > in the command line, prefix kwargs with '-',
                # optional kwargs with '--'
                if name in hopla_optional:
                    option = ["--" + name]
                else:
                    option = ["-" + name]
                # > set the option value
                if name in append_kwargs:
                    _option_key = option[0]
                    option = []
                    for elem in value:
                        option.append(_option_key)
                        if isinstance(elem, list):
                            option.extend(elem)
                        else:
                            option.append(elem)
                elif isinstance(value, list):
                    option.extend([str(item) for item in value])
                elif not isinstance(value, bool):
                    option.append(str(value))
                # > set the option in the command line: for boolean set the
                # option only if the value is true
                if isinstance(value, bool):
                    if value:
                        cmd.extend(option)
                else:
                    cmd.extend(option)

    # Execute the commands with a scheduler in order to control the execution
    # load
    script_name = python_script.split(os.sep)[-1].split(".")[0]
    return scheduler(commands=commands,
                     name=script_name,
                     outputdir=hopla_outputdir,
                     cpus=hopla_cpus,
                     use_subprocess=hopla_use_subprocess,
                     delay_upto=hopla_delay_upto,
                     logfile=hopla_logfile,
                     cluster=hopla_cluster,
                     cluster_logdir=hopla_cluster_logdir,
                     cluster_queue=hopla_cluster_queue,
                     cluster_memory=hopla_cluster_memory,
                     cluster_walltime=hopla_cluster_walltime,
                     cluster_nb_threads=hopla_cluster_nb_threads,
                     python_cmd=hopla_python_cmd,
                     verbose=hopla_verbose)



##########################################################################
# Hopla - Copyright (C) AGrigis, 2015 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module proposes a local worker and a distant TORQUE worker. The two
proposed workers are able to follow a '__hopla__' list of parameter
names to keep trace on. All specified parameters values are stored in the
execution status.
"""  # pragma: no cover

# System import
from __future__ import print_function  # pragma: no cover
import os  # pragma: no cover
import copy  # pragma: no cover
import subprocess  # pragma: no cover
import traceback  # pragma: no cover
from socket import getfqdn  # pragma: no cover
import sys  # pragma: no cover
import glob  # pragma: no cover
import time  # pragma: no cover
import json  # pragma: no cover
import random  # pragma: no cover

# Hopla import
from .signals import FLAG_ALL_DONE  # pragma: no cover
from .signals import FLAG_WORKER_FINISHED_PROCESSING  # pragma: no cover


def worker(tasks, returncodes, python_cmd="python", delay_upto=0,
           use_subprocess=False):
    """ The worker function for a script.

    If the script contains a '__hopla__' list of parameter names to keep
    trace on, all the specified parameters values are stored in the return
    code.

    Parameters
    ----------
    tasks, returncodes: multiprocessing.Queue
        the input (commands) and output (results) queues.
    python_cmd: str (optional, default 'python')
        the path to the python binary. Only usefull in the subprocess mode.
    delay_upto: int (optional, default 0)
        the process execution will be delayed randomly by [0, <delay_upto>[
        seconds.
    use_subprocess: bool, default False
        use a subprocess (for instance in case of memory leak). In this
        particular case the __hopla__ setting is deactivated.
    """
    while True:
        signal = tasks.get()
        if signal == FLAG_ALL_DONE:
            returncodes.put(FLAG_WORKER_FINISHED_PROCESSING)
            break
        job_name, command = signal
        returncode = {}
        returncode[job_name] = {}
        returncode[job_name]["info"] = {}
        returncode[job_name]["debug"] = {}
        returncode[job_name]["info"]["cmd"] = command
        returncode[job_name]["debug"]["hostname"] = getfqdn()

        # COMPATIBILITY: dict in python 2 becomes structure in pyton 3
        python_version = sys.version_info
        if python_version[0] < 3:
            environ = copy.deepcopy(os.environ.__dict__)
        else:
            environ = copy.deepcopy(os.environ._data)
        returncode[job_name]["debug"]["environ"] = environ

        # Execution with a random delay expressed in seconds
        try:
            time.sleep(random.random() * abs(delay_upto))
            sys.argv = command
            job_status = {}
            if use_subprocess:
                if python_cmd is not None:
                    subprocess.check_call([python_cmd] + command)
                else:
                    subprocess.check_call(command)
            else:
                with open(command[0]) as ofile:
                    exec(ofile.read(), job_status)
            returncode[job_name]["info"]["exitcode"] = "0"
        # Error
        except:
            returncode[job_name]["info"]["exitcode"] = (
                "1 - '{0}'".format(traceback.format_exc().rstrip("\n")))
        # Follow '__hopla__' script parameters
        finally:
            if "__hopla__" in job_status:
                for parameter_name in job_status["__hopla__"]:
                    if parameter_name in job_status:
                        returncode[job_name]["info"][
                            parameter_name] = job_status[parameter_name]
        returncodes.put(returncode)


PBS_TEMPLATE = """
#!/bin/bash
#PBS -l mem={memory}gb,nodes=1:ppn={threads},walltime={hwalltime}:00:00
#PBS -N {name}
#PBS -e {errfile}
#PBS -o {logfile}
echo $PBS_JOBID
{command}
"""  # pragma: no cover


PY_TEMPLATE = """
from __future__ import print_function
import sys
import json


# Execute the command line in the 'job_status' environment
try:
    command = {cmd}
    sys.argv = command
    job_status = dict()
    with open(command[0]) as ofile:
        exec(ofile.read(), job_status)
# Error
except:
    raise
# Follow '__hopla__' script parameters: print the parameters to keep trace on
# in '<hopla>...</hopla>' div in order to communicate with the scheduler and
# in order to generate a complete log
finally:
    parameters = dict()
    if "__hopla__" in job_status:
        for parameter_name in job_status["__hopla__"]:
            if parameter_name in job_status:
                parameters[parameter_name] = job_status[parameter_name]
    print("<hopla>")
    print(json.dumps(parameters))
    print("</hopla>")
"""  # pragma: no cover


def qsub_worker(tasks, returncodes, logdir, queue,
                memory=1, walltime=24, nb_threads=1, python_cmd="python",
                delay_upto=0, sleep=4):
    """ A cluster worker function for a script.

    Use the TORQUE resource manager provides control over batch jobs and
    distributed computing resources. It is an advanced open-source product
    based on the original PBS project.

    Use a double script strategy in order to manage the '__hopla__' list of
    parameter names to keep trace on: a '.pbs' script calling another '.py'
    script that print the '__hopla__' parameters. All the specified parameters
    values are stored in the return code.

    Parameters
    ----------
    tasks, returncodes: multiprocessing.Queue
        the input (commands) and output (results) queues.
    logdir: str
        a path where the qsub error and output files will be stored.
    queue: str
        the name of the queue where the jobs will be submited.
    memory: float (optional, default 1)
        the memory allocated to each qsub (in GB).
    walltime: int (optional, default 24)
        the walltime used for each job submitted on the cluster (in hours).
    nb_threads: int (optional, default 1)
        the number of cores allocated for each node.
    python_cmd: str (optional, default 'python')
        the path to the python binary. If None consider the command directly in
        the PBS batch.
    delay_upto: int (optional, default 0)
        the process execution will be delayed randomly by [0, <delay_upto>[
        seconds.
    sleep: float (optional, default 4)
        time rate to check the termination of the submited jobs.
    """
    while True:
        signal = tasks.get()
        if signal == FLAG_ALL_DONE:
            returncodes.put(FLAG_WORKER_FINISHED_PROCESSING)
            break
        job_name, command = signal
        returncode = {}
        returncode[job_name] = {}
        returncode[job_name]["info"] = {}
        returncode[job_name]["debug"] = {}
        returncode[job_name]["info"]["cmd"] = command
        returncode[job_name]["debug"]["hostname"] = getfqdn()

        # COMPATIBILITY: dict in python 2 becomes structure in python 3
        python_version = sys.version_info
        if python_version[0] < 3:
            environ = copy.deepcopy(os.environ.__dict__)
        else:
            environ = copy.deepcopy(os.environ._data)
        returncode[job_name]["debug"]["environ"] = environ

        # Torque-PBS execution
        fname_pbs = os.path.join(logdir, job_name + ".pbs")
        fname_py = os.path.join(logdir, job_name + ".py")
        errfile = os.path.join(logdir, "error." + job_name)
        logfile = os.path.join(logdir, "output." + job_name)
        try:
            # Random delay expressed in seconds
            time.sleep(random.random() * abs(delay_upto))

            # Edit the job to be submitted
            if python_cmd is not None:
                with open(fname_py, "w") as open_file:
                    open_file.write(PY_TEMPLATE.format(cmd=command))
                with open(fname_pbs, "w") as open_file:
                    pbs_cmd = " ".join([python_cmd, fname_py])
                    open_file.write(PBS_TEMPLATE.format(
                        memory=memory,
                        hwalltime=walltime,
                        threads=nb_threads,
                        name=job_name,
                        errfile=errfile,
                        logfile=logfile,
                        command=pbs_cmd))
            else:
                with open(fname_pbs, "w") as open_file:
                    open_file.write(PBS_TEMPLATE.format(
                        memory=memory,
                        hwalltime=walltime,
                        threads=nb_threads,
                        name=job_name,
                        errfile=errfile,
                        logfile=logfile,
                        command=" ".join(command)))

            # Submit the job
            # subprocess.check_call(["qsub", "-q", queue, fname_pbs])
            process = subprocess.Popen(["qsub", "-q", queue, fname_pbs],
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
            stdout, stderr = process.communicate()
            job_id = stdout.rstrip("\n")
            print(job_id)
            exitcode = process.returncode
            if exitcode != 0:
                raise Exception(stderr)

            # Lock everything until the submitted command has not terminated
            while True:
                terminated = (len(glob.glob(errfile + ".*")) > 0 or
                              len(glob.glob(logfile + ".*")) > 0)
                with_log = terminated
                process = subprocess.Popen("qstat | grep {0}".format(job_id),
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE,
                                           shell=True)
                stdout, stderr = process.communicate()
                exitcode = process.returncode
                terminated = terminated or (exitcode == 1)
                if terminated:
                    break
                time.sleep(sleep)

            # Check that no error was produced during the submission
            if with_log:
                with open(glob.glob(errfile + ".*")[0]) as open_file:
                    stderr = open_file.readlines()
                if len(stderr) > 0:
                    raise Exception("\n".join(stderr))

            # Update the return code
            if with_log:
                returncode[job_name]["info"]["exitcode"] = "0"
            else:
                returncode[job_name]["info"]["exitcode"] = "-1"
        # Error
        except:
            if os.path.isfile(errfile):
                with open(errfile) as openfile:
                    error_message = openfile.readlines()
            else:
                error_message = traceback.format_exc()
            returncode[job_name]["info"]["exitcode"] = (
                "1 - '{0}'".format(error_message))
        # Follow '__hopla__' script parameters in pbs '<hopla>...</hopla>'
        # output
        finally:
            pbs_logfiles = glob.glob(logfile + ".*")
            if len(pbs_logfiles) == 1:
                with open(pbs_logfiles[0]) as open_file:
                    stdout = open_file.read()
                hopla_start = stdout.rfind("<hopla>")
                hopla_end = stdout.rfind("</hopla>")
                parameters_repr = stdout[
                    hopla_start + len("<hopla>"): hopla_end]
                try:
                    parameters = json.loads(parameters_repr)
                except:
                    parameters = {}
                for name, value in parameters.items():
                    returncode[job_name]["info"][name] = value

        returncodes.put(returncode)



#! /usr/bin/env python
##########################################################################
# Hopla - Copyright (C) AGrigis, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# Hopla globals
from .info import DESCRIPTION, __version__  # pragma: no cover
from .scheduler import scheduler  # pragma: no cover
from .converter import hopla  # pragma: no cover

__doc__ = DESCRIPTION  # pragma: no cover



##########################################################################
# Hopla - Copyright (C) AGrigis, 2015 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module proposes a simple scheduler used to execute a list of tasks
controlling the machine load.
"""  # pragma: no cover

# System import
import os  # pragma: no cover
import logging  # pragma: no cover
import json  # pragma: no cover
import multiprocessing  # pragma: no cover

# Hopla import
import hopla  # pragma: no cover
from .workers import worker  # pragma: no cover
from .workers import qsub_worker  # pragma: no cover
from .signals import FLAG_ALL_DONE  # pragma: no cover
from .signals import FLAG_WORKER_FINISHED_PROCESSING  # pragma: no cover

# Define the logger for this file
multiprocessing.log_to_stderr(logging.CRITICAL)  # pragma: no cover
logger = logging.getLogger("hopla")  # pragma: no cover


def scheduler(commands, name="job", outputdir=None, cpus=1,
              use_subprocess=False, delay_upto=0,
              logfile=None, cluster=False, cluster_logdir=None,
              cluster_queue=None, cluster_memory=1, cluster_walltime=24,
              cluster_nb_threads=1, python_cmd="python", verbose=1):
    """ Execute some commands (python scripts) using a scheduler.

    If the script contains a '__hopla__' list of parameter names to keep
    trace on, all the specified parameters values are stored in the execution
    status.

    Parameters
    ----------
    commands: list of list of str (mandatory)
        some commands to be executed: the first command element must be a
        path to a python script.
    name: str (optional, default 'job')
        the job name is constructed with this parameter as <name>_<iter>.
    outputdir: str (optional, default None)
        a folder where a summary of the executed jobs are written.
    cpus: int (optional, default 1)
        the number of cpus to be used.
    use_subprocess: bool, default False
        use a subprocess (for instance in case of memory leak). In this
        particular case the __hopla__ setting is deactivated.
    delay_upto: int (optional, default 0)
        the processes' execution will be delayed randomly by [0, <delay_upto>[
        seconds.
    logfile: str (optional, default None)
        location where the log messages are redirected: INFO and DEBUG.
    cluster: bool (optional, default False)
        if True use a worker that submits the jobs to a cluster.
    cluster_logdir: str (optional, default None)
        an existing path where the cluster error and output files will be
        stored. This folder must be empty.
    cluster_queue: str (optional, default None)
        the name of the queue where the jobs will be submited.
    cluster_memory: float (optional, default 1)
        the memory allocated to each job submitted on a cluster (in GB).
    cluster_walltime: int (optional, default 24)
        the walltime used for each job submitted on the cluster (in hours).
    cluster_nb_threads: int (optional, default 1)
        the number of cores allocated for each node.
    python_cmd: str (optional, default 'python')
        the path to the python binary. May also be used in subprocess mode.
    verbose: int (optional, default 1)
        0 - display no log in console,
        1 - display information log in console,
        2 - display debug log in console.

    Returns
    -------
    execution_status: dict
        a dictionary that contains all the executed command return codes.
    exitcodes: dict
        a dictionary with a summary of the executed jobs exit codes.
    """
    # If someone tried to log something before basicConfig is called,
    # Python creates a default handler that goes to the console and
    # will ignore further basicConfig calls: we need to remove the
    # handlers if there is one.
    while len(logging.root.handlers) > 0:
        logging.root.removeHandler(logging.root.handlers[-1])

    # Remove console and file handlers if already created
    while len(logger.handlers) > 0:
        logger.removeHandler(logger.handlers[-1])

    # Create console handler.
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    if verbose != 0:
        console_handler = logging.StreamHandler()
        if verbose == 1:
            logger.setLevel(logging.INFO)
            console_handler.setLevel(logging.INFO)
        else:
            logger.setLevel(logging.DEBUG)
            console_handler.setLevel(logging.DEBUG)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    else:
        logger.setLevel(logging.INFO)
        logger.addHandler(logging.NullHandler())

    # Create a file handler if requested
    if logfile is not None:
        file_handler = logging.FileHandler(logfile, mode="a")
        if verbose > 1:
            file_handler.setLevel(logging.DEBUG)
        else:
            file_handler.setLevel(logging.INFO)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        logger.info("Processing information will be logged in file "
                    "'{0}'.".format(logfile))

    # Information
    logger.info("Using 'hopla' version '{0}'.".format(hopla.__version__))
    exit_rules = [
        "For exitcode values:",
        "    = 0 - no error was produced.",
        "    > 0 - the process had an error, and exited with that code.",
        "    < 0 - the process was killed with a signal of -1 * exitcode."]
    logger.info("\n".join(exit_rules))

    # Get the machine available cpus for local processings
    if not cluster:
        nb_cpus = multiprocessing.cpu_count() - 1
        nb_cpus = nb_cpus or 1
        if max(cpus, nb_cpus) == cpus:
            cpus = nb_cpus

    # Create the workers
    # Works as a FIFO with 1 cpu
    workers = []
    tasks = multiprocessing.Queue()
    returncodes = multiprocessing.Queue()
    for index in range(cpus):
        if cluster:
            if cluster_logdir is None or not os.path.isdir(cluster_logdir):
                raise ValueError(
                    "'{0}' is not a valid directory to set cluster "
                    "logs.".format(cluster_logdir))
            if len(os.listdir(cluster_logdir)) > 0:
                raise ValueError(
                    "'{0}' is not an empty directory.".format(cluster_logdir))
            if cluster_queue is None:
                raise ValueError(
                    "In order to execute the jobs on the cluster, please "
                    "specify an execution queue.")
            process = multiprocessing.Process(
                target=qsub_worker, args=(tasks, returncodes, cluster_logdir,
                                          cluster_queue, cluster_memory,
                                          cluster_walltime, cluster_nb_threads,
                                          python_cmd, delay_upto))
        else:
            process = multiprocessing.Process(
                target=worker, args=(tasks, returncodes, python_cmd,
                                     delay_upto, use_subprocess))
        process.deamon = True
        process.start()
        workers.append(process)

    # Execute the input commands
    # Use a FIFO strategy to deal with multiple boxes
    execution_status = {}
    workers_finished = 0
    try:
        # Assert something has to be executed
        if len(commands) == 0:
            raise ValueError("Nothing to execute.")

        # Add all the jobs to the 'tasks' queue
        for cnt, cmd in enumerate(commands):
            job_name = "{0}_{1}".format(name, cnt)
            tasks.put((job_name, cmd))

        # Add poison pills to stop the remote workers
        for index in range(cpus):
            tasks.put(FLAG_ALL_DONE)

        # Loop until all the jobs are finished
        while True:

            # Collect the box returncodes
            wave_returncode = returncodes.get()
            if wave_returncode == FLAG_WORKER_FINISHED_PROCESSING:
                workers_finished += 1
                if workers_finished == cpus:
                    break
                continue
            job_name = list(wave_returncode.keys())[0]
            execution_status.update(wave_returncode)

            # Information
            for key, value in wave_returncode[job_name]["info"].items():
                logger.info("{0}.{1} = {2}".format(
                    job_name, key, value))
            for key, value in wave_returncode[job_name]["debug"].items():
                logger.debug("{0}.{1} = {2}".format(
                    job_name, key, value))
    except:
        # Stop properly all the workers before raising the exception
        for process in workers:
            process.terminate()
            process.join()
        raise

    # Save processing status to ease the generated data interpretation if the
    # 'outputdir' is not None
    exitcodes = {}
    for job_name, job_returncode in execution_status.items():
        exitcodes[job_name] = int(job_returncode["info"]["exitcode"].split(
            " - ")[0])
    if outputdir is not None:
        exitcodes_file = os.path.join(outputdir, "scheduler_status.json")
        with open(exitcodes_file, "w") as open_file:
            json.dump(exitcodes, open_file, indent=4, sort_keys=True)

    return execution_status, exitcodes



#! /usr/bin/env python
##########################################################################
# Hopla - Copyright (C) AGrigis, 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# pragma: no cover

# Pysphinxdoc current version
version_major = 1  # pragma: no cover
version_minor = 0  # pragma: no cover
version_micro = 5  # pragma: no cover

# Expected by setup.py: string of form "X.Y.Z"
__version__ = "{0}.{1}.{2}".format(
    version_major, version_minor, version_micro)  # pragma: no cover

# Project descriptions
DESCRIPTION = "[hopla] Easy to use pure-Python scheduler."  # pragma: no cover
LONG_DESCRIPTION = """
Easy to use pure-Python scheduler.

Overview
========

With the increasing amount of data to be treated, efficient scaling strategies
are necessary. This observation made me start 'hopla' which provides:

- a scheduler that produces human readable outputs.
- a converter that enables to execute kilometer command lines
- workers that enable local or cluster executions.


Usage
=====

The proposed module has been currently developped to execute Python scripts.
Consider the following demonstration script that lists an input folder (the
latter is available in the 'hopla.demo.my_ls_script' module)::

    #! /usr/bin/env python
    ##########################################################################
    # Hopla - Copyright (C) AGrigis, 2015
    # Distributed under the terms of the CeCILL-B license, as published by
    # the CEA-CNRS-INRIA. Refer to the LICENSE file or to
    # http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
    # for details.
    ##########################################################################

    # Bredala import
    try:
        import bredala
        bredala.USE_PROFILER = False
        bredala.register("os.path", names=["listdir"])
    except:
        pass

    # System import
    from datetime import datetime
    import argparse
    import os

    # Parameters to keep trace
    __hopla__ = ["runtime", "inputs", "outputs"]


    def is_directory(dirarg):
        if not os.path.isdir(dirarg):
            raise argparse.ArgumentError(
                "The directory '{0}' does not exist!".format(dirarg))
        return dirarg


    parser = argparse.ArgumentParser(description="List a directory.")
    required = parser.add_argument_group("required arguments")
    required.add_argument(
        "-d", "--dir", dest="dir", required=True, metavar="PATH",
        help="a valid directory to be listed.", type=is_directory)
    parser.add_argument(
        "-b", "--fbreak", dest="fbreak", action="store_true",
        help="a activated raise a ValueError.")
    required.add_argument(
        "-l", "--mylist", dest="mylist", nargs="+", type=int, required=True,
        help="a list that will be printed.")
    parser.add_argument(
        "-v", "--verbose", dest="verbose", type=int, choices=[0, 1, 2],
        default=0,
        help="increase the verbosity level: 0 silent, [1, 2] verbose.")
    args = parser.parse_args()

    directory = args.dir
    break_flag = args.fbreak
    mylist = args.mylist
    runtime = {
        "timestamp": datetime.now().isoformat()
    }
    inputs = {
        "directory": directory,
        "break_flag": break_flag,
        "mylist": mylist
    }
    outputs = None
    if break_flag:
        raise ValueError("BREAK ACTIVATED.")
    files = os.listdir(directory)
    print("[res] --------", mylist)
    if args.verbose > 0:
        print("[res] --------", files)
    outputs = {
        "files": files
    }

Note the '__hopla__' list that specifies which parameters will be dispalyed in
the scheduler execution log. This mechanism is usefull to keep trace of
important script elements. The scaled execution of this script on two CPUs is
realized using a simple call::

    # System import
    from pprint import pprint

    # Hopla import
    from hopla.converter import hopla
    import hopla as root

    # Define script parameters
    apath = os.path.join(os.path.abspath(os.path.dirname(root.__file__)),
                         "demo")
    script = os.path.join(apath, "my_ls_script.py")

    # Local execution
    status, exitcodes = hopla(
        script, hopla_iterative_kwargs=["d", "b"], verbose=0, l=[1, 2],
        b=[False, True, False], d=[apath, apath, apath], hopla_verbose=1,
        hopla_cpus=2)
    pprint(status)
    pprint(exitcodes)

After the execution call (through the hopla function), exit codes are
inspected. The 'hopla_verbose' has been set to one, some logging information
has been displayed::

    2016-08-02 15:59:26,562 - INFO - Using 'hopla' version '1.0.2'.
    2016-08-02 15:59:26,562 - INFO - For exitcode values:
        = 0 - no error was produced.
        > 0 - the process had an error, and exited with that code.
        < 0 - the process was killed with a signal of -1 * exitcode.
    2016-08-02 15:59:26,927 - INFO - job_0.inputs = {
        'directory': '/home/ag239446/git/hopla/hopla/demo',
        'break_flag': False,
        'mylist': [1, 2]}
    2016-08-02 15:59:26,928 - INFO - job_0.exitcode = 0
    2016-08-02 15:59:26,928 - INFO - job_0.cmd = [
        '/home/ag239446/git/hopla/hopla/demo/my_ls_script.py', '--dir',
        '/home/ag239446/git/hopla/hopla/demo', '--mylist', '1', '2',
        '--verbose', '1']
    2016-08-02 15:59:26,928 - INFO - job_0.outputs = {
        'files': ['my_ls_script.py', 'demo.py']}
    2016-08-02 15:59:26,928 - INFO - job_0.runtime = {
        'timestamp': '2016-08-02T15:59:26.926153'}
    2016-08-02 15:59:26,928 - INFO - job_1.inputs = {
        'directory': '/home/ag239446/git/hopla/hopla/demo',
        'break_flag': True,
        'mylist': [1, 2]}
    2016-08-02 15:59:26,929 - INFO - job_1.exitcode = 1 - 'Traceback (most
      recent call last):
      File "/home/ag239446/git/hopla/hopla/workers.py", line 70, in worker
        exec(ofile.read(), job_status)
      File "<string>", line 65, in <module>
    ValueError: BREAK ACTIVATED.
    '
    2016-08-02 15:59:26,929 - INFO - job_1.cmd = [
        '/home/ag239446/git/hopla/hopla/demo/my_ls_script.py', '-b', '--dir',
        '/home/ag239446/git/hopla/hopla/demo', '--mylist', '1', '2',
        '--verbose', '1']
    2016-08-02 15:59:26,929 - INFO - job_1.outputs = None
    2016-08-02 15:59:26,929 - INFO - job_1.runtime = {
        'timestamp': '2016-08-02T15:59:26.926772'}
    2016-08-02 15:59:26,979 - INFO - job_2.inputs = {
        'directory': '/home/ag239446/git/hopla/hopla/demo',
        'break_flag': False,
        'mylist': [1, 2]}
    2016-08-02 15:59:26,979 - INFO - job_2.exitcode = 0
    2016-08-02 15:59:26,979 - INFO - job_2.cmd = [
        '/home/ag239446/git/hopla/hopla/demo/my_ls_script.py', '--dir',
        '/home/ag239446/git/hopla/hopla/demo', '--mylist', '1', '2',
        '--verbose', '1']
    2016-08-02 15:59:26,979 - INFO - job_2.outputs = {
        'files': ['my_ls_script.py', 'demo.py']}
    2016-08-02 15:59:26,979 - INFO - job_2.runtime = {
        'timestamp': '2016-08-02T15:59:26.969334'}


Perspectives
============

It will be nice to generalize some concepts (ie., accept different kind
of scripts).
"""  # pragma: no cover

# Main setup parameters
NAME = "hopla"  # pragma: no cover
AUTHOR = "Antoine Grigis"  # pragma: no cover
AUTHOR_EMAIL = "antoine.grigis@cea.fr"  # pragma: no cover
URL = "https://github.com/AGrigis/hopla"  # pragma: no cover
DOWNLOAD_URL = "https://github.com/AGrigis/hopla"  # pragma: no cover
LICENSE = "GPL 2+"  # pragma: no cover
VERSION = __version__  # pragma: no cover



#! /usr/bin/env python
##########################################################################
# Bredala - Copyright (C) AGrigis, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# System import
from setuptools import setup, find_packages
from os.path import join, dirname
import hopla
import sys


description = open(join(dirname(__file__), "README.rst")).read()
pkgdata = {
    "hopla.test": ["*.py"],
}

setup(
    name="hopla",
    version=hopla.__version__,
    author="Antoine Grigis",
    author_email="antoine.grigis@cea.fr",
    description=next(x for x in description.splitlines() if x.strip()),
    long_description=".. contents::\n\n" + description,
    url="http://github.com/AGrigis/hopla",
    license="GPL 2+",
    platforms=["any"],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v2 or "
        "later (GPLv2+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development",
    ],
    packages=find_packages(),
    package_data=pkgdata,
    install_requires=[]
)



#! /usr/bin/env python
##########################################################################
# Hopla - Copyright (C) AGrigis, 2015 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the scheduler signals.
"""  # pragma: no cover

# Define signals
FLAG_ALL_DONE = b"WORK_FINISHED"  # pragma: no cover
FLAG_WORKER_FINISHED_PROCESSING = (
    b"WORKER_FINISHED_PROCESSING")  # pragma: no cover
